#include "HelixPso.h"

using namespace std;


//static int TP=0;

double NO_ENERGY_IMPROVEMENT=10000.0;
//double sol_generated_count=0;
int GREEDY=0;
int KAWASAKI=1;
int METROPOLIS=2;
///XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX RAUS
int matched;
int record=0;

double HelixPso::alpha;

double HelixPso::sim_anneal_start;
double HelixPso::sim_anneal_end;
 int HelixPso::iterations_done;
 int HelixPso::sv_stat_sections;
 std::vector<int> HelixPso::sv_length_counter;
 std::vector<double> HelixPso::sv_mass_counter;
 std::vector<std::vector<double> > HelixPso::sv_distribution_counter;
 std::vector<double> HelixPso::sv_mass_fractions;

 std::vector<std::vector<int> >  HelixPso::transposition1st_counter;
 std::vector<std::vector<int> >  HelixPso::transposition2nd_counter;
 int HelixPso::transposition_count_sections;

 std::vector<int>  HelixPso::accepted_solution_counter;
 int HelixPso::accepted_solution_count_sections;

std::vector<std::vector<std::pair<int,int> > > HelixPso::stemlist;


double HelixPso::evaporation_threshold;
double HelixPso::swap_select;

double HelixPso::centroid_weight;


std::vector<std::pair<int,int> >  HelixPso::bp_native;
std::vector<std::pair<int,int> >  HelixPso::bp_centroid;


int HelixPso::rule;
std::pair<int,int> HelixPso::best_pred_in_all_runs= std::pair<int,int>(0,0);
std::pair<int,int> HelixPso::best_lastgen= std::pair<int,int>(0,0);


int  HelixPso::RandInt(int max){  
  return (int) (RandDouble()*max);
}

double  HelixPso::RandDouble(){
 return drand48();
}







std::vector<int>  HelixPso::GetRandomPermutation(int n){
  vector<int> ret=vector<int>();
  vector<int> pool=vector<int>();
  for(int i=0;i<n;i++) pool.push_back(i);

  for(int i=0;i<n-1;i++){
    int idx=RandInt(pool.size());
    int val=pool[idx];   
    ret.push_back(val);
    vector<int>::iterator it=find(pool.begin(),pool.end(),val);
    pool.erase(it);
    
  }
  //add the remaining element
  ret.push_back(pool.front());
  return ret;

}

std::vector<std::pair<int,int> > HelixPso::ChromosomeToNode(const vector<int> &chromosome,const std::vector<std::vector<std::pair<int,int> > >& stemlist){
  
  std::vector<std::pair<int,int> > node=std::vector<std::pair<int,int> >();
  node.reserve(300);
  for(size_t i=0;i<chromosome.size();i++){
    int helix_idx=chromosome[i];
    std::vector<std::pair<int,int> > helix=stemlist[helix_idx];
    if(!Swarm::Conflict(node,helix)){ 
      for(size_t j=0;j<stemlist[helix_idx].size();j++) node.push_back(stemlist[helix_idx][j]);
    }
  }
  return node;
}


HelixPso::HelixPso(RNAParameters* p,SearchParameters* s,int clusters){
  gbest_changed=true;
  this->p=p;
  this->s=s;
  last_stat_update=0;
  sv_stat_section=0;
  transposition_count_section=0;
  accepted_solution_count_section=0;

  sim_anneal=sim_anneal_start;
  solution_count=0;
  this->total_solutions=s->total_solutions;
  this->solutions_report_interval=s->solutions_report_interval;
  solutions_report=std::vector<double>();
  sequence=p->RNA;
  population=s->ants;
  iterations=s->iterations;
  HelixPso::iterations_done=0;
  this->clusters=clusters;

  chromosomes= std::vector<std::pair<double,std::vector<int> > >() ;
  nodes=std::vector<std::vector<std::pair<int,int> > > ();
  pbest = std::vector<std::pair<double,std::vector<int> > >();
  gbest= std::vector<std::pair<double,std::vector<int> > >() ;
  clbest=  std::vector<std::vector<std::pair<double,std::vector<int> > > >() ;
  for(int i=0;i<clusters;i++) clbest.push_back(std::vector<std::pair<double,std::vector<int> > >());
  SV=std::vector<std::map<std::pair<double,std::vector<int> >,double> > ();
  SetStartAndStopIdx();
 
  //cout<<" Algo settings gbest "+Str(s->p_bonus_gbest)+" pbest "+Str(s->p_bonus_pbest)+" random "+Str(s->p_random_move)+" rule "+Str(HelixPso::rule)<<std::endl;
}

std::vector<std::string> HelixPso::GetStructures(){
  std::vector<std::string> structures=std::vector<std::string>();
  for(int i=0;i<(int)nodes.size();i++){
    std::string struc=RNAParameters::Structure(p->RNA.size(),nodes[i]);
    structures.push_back(struc);
  }
  return structures;
}


int HelixPso::Distance(std::string str1,std::string str2){
  std::vector<std::pair<int,int> > bp_str1=RNAParameters::MakeBasePairList(str1);
  std::vector<std::pair<int,int> > bp_str2=RNAParameters::MakeBasePairList(str2);

  int bpcorrect=0;
  for(std::vector<std::pair<int,int> >::iterator it=bp_str1.begin();it!=bp_str1.end();it++){
        if(find(bp_str2.begin(),bp_str2.end(),*it)!= bp_str2.end()) bpcorrect++;
  }
  //int dist=bp_str1.size()-bpcorrect;
    int dist=bp_str1.size()+bp_str2.size()-bpcorrect;
  return dist;

}

std::vector<std::vector<int> > CalculateDistanceMatrix(std::vector<std::string> structures){
  std::vector<std::vector<int> > distance=std::vector<std::vector<int> >  ();

  for(int i=0;i<(int)structures.size();i++){
    std::vector<int> this_distance=std::vector<int> ();
    for(int j=0;j<(int)structures.size();j++) this_distance.push_back(HelixPso::Distance(structures[i],structures[j]));
    distance.push_back(this_distance);
  }
  return distance;
}

int Sum(std::vector<int> v){
  int sum=0;
  for(int i=0;i<(int)v.size();i++) sum+=v[i];
  return sum;
}


std::string HelixPso::SelectLowestDistanceStructure(){
  std::vector<std::string> structures=GetStructures();
  std::vector<std::vector<int> > distance=CalculateDistanceMatrix(structures);


  std::vector<int> distance_sum=std::vector<int> ();
  for(int i=0;i<(int)structures.size();i++){
    int sum=Sum(distance[i]);
    distance_sum.push_back(sum);
    cout<<structures[i]<<" dist "+Str(sum)<<std::endl;
  }

  int min_idx=min_element(distance_sum.begin(),distance_sum.end())-distance_sum.begin();
  std::string structure=structures[min_idx];
  double energy=energy_of_struct(const_cast<char*>(p->RNA.c_str()),const_cast<char*>(structure.c_str()));
     int correct=CorrectlyPredictedBpCount(structure).first;
     cout<<"min distance "+structure+" "+Str(energy)+" "+Str(correct)+" "+Str(distance_sum[min_idx])+" at idx "+Str(min_idx)<<std::endl;
  return structures[min_idx];
}


int HelixPso::StructureHasFrequency(std::string structure){

std::vector<std::string> structures=GetStructures();
 int count=0;
 for(int i=0;i<(int)structures.size();i++){
   if(structures[i]==structure) count++;
 }
 return count;
}

std::vector<std::string> HelixPso::ReturnMostFrequentStructures(){
    bool verbose=false;
  std::vector<int> frequency=std::vector<int>();
  std::vector<std::string> structures=GetStructures();
 std::vector<std::string> frequent_structures= std::vector<std::string>();
   std::vector<std::string> unique_structures=std::vector<std::string> ();
   for(int i=0;i<(int)structures.size();i++){
     std::vector<std::string>::iterator it=find(unique_structures.begin(),unique_structures.end(),structures[i]);
     int idx=it-unique_structures.begin();
     if(it==unique_structures.end()){
       unique_structures.push_back(structures[i]);
       frequency.push_back(1);
     }
     else frequency[idx]++;
   }
   //find most often occurring node
   int max_idx=max_element(frequency.begin(),frequency.end())-frequency.begin();
   int max_count=frequency[max_idx];
   //  std::vector<int> max_idxs=std::vector<int> ();
   for(int i=0;i<(int)frequency.size();i++){
     if(frequency[i]==max_count)  frequent_structures.push_back(unique_structures[i]);
   }


    if(verbose){
   cout<<"frequencies "<<std::endl;
   for(int j=0;j<(int)frequency.size();j++){
          if(frequency[j]>1 && frequency[j]>=.5*frequency[max_idx]){
     std::string struc=unique_structures[j];//RNAParameters::Structure(p->RNA.size(),unique_nodes[j]);
     double energy=energy_of_struct(const_cast<char*>(p->RNA.c_str()),const_cast<char*>(struc.c_str()));
     int correct=CorrectlyPredictedBpCount(struc).first;
     cout<<Str(j)+" "+struc+" "+Str(energy)+" "+Str(correct)+" freq "+Str(frequency[j])<<std::endl;
     	  }
   }
   
   cout<<"max frequency "+Str(frequency[max_idx])+" at idx "+Str(max_idx)<<std::endl;
   }
   // return unique_structures[max_idx];
   return frequent_structures;
}


void HelixPso::InitializeTranspositionCounters(){ 
  transposition1st_counter= std::vector<std::vector<int> >  ();
  transposition2nd_counter= std::vector<std::vector<int> >  ();

  if(HelixPso::transposition_count_sections>0){
    transposition1st_counter.reserve(HelixPso::transposition_count_sections);
    transposition2nd_counter.reserve(HelixPso::transposition_count_sections);
    for(int i=0;i<HelixPso::transposition_count_sections;i++){
	transposition1st_counter.push_back(std::vector<int>(HelixPso::stemlist.size(),0));
	transposition2nd_counter.push_back(std::vector<int>(HelixPso::stemlist.size(),0));
	}
  }
}


void HelixPso::InitializeAcceptedSolutionCounter(){ 
  if(HelixPso::accepted_solution_count_sections>0){
    accepted_solution_counter=std::vector<int>(HelixPso::accepted_solution_count_sections,0);
  }
}
void  HelixPso::PrepareSVStatistics(){
 
  if(HelixPso::sv_stat_sections>0){
    HelixPso::sv_mass_fractions=std::vector<double>();
    //important:fill greatest to smallest. The > comparison when doing statistics in UpdateSV relies on it
    HelixPso::sv_mass_fractions.push_back(.1);
    HelixPso::sv_mass_fractions.push_back(.05);
    HelixPso::sv_mass_fractions.push_back(.02);
    HelixPso::sv_mass_fractions.push_back(.01);
    HelixPso::sv_mass_fractions.push_back(.005);
    HelixPso::sv_mass_fractions.push_back(.001);

    HelixPso::sv_distribution_counter=std::vector<std::vector<double> >  ();
    HelixPso::sv_distribution_counter.reserve(HelixPso::sv_stat_sections);
    for(int i=0;i<HelixPso::sv_stat_sections;i++) sv_distribution_counter.push_back(std::vector<double>(HelixPso::sv_mass_fractions.size(),0.0));

    HelixPso::sv_length_counter=std::vector<int>(HelixPso::sv_stat_sections,0);
    HelixPso::sv_mass_counter=std::vector<double>(HelixPso::sv_stat_sections,0.0);
  }
}

void HelixPso::InitializeControlVariables(){
  InitializeSV();
  pbest=chromosomes;
  UpdateClbest();
  UpdateGbest();
}



void HelixPso::CreateStemList(RNAParameters p){
  HelixPso::stemlist=p.EnumerateStems();
  sort(HelixPso::stemlist.begin(),HelixPso::stemlist.end(),Swarm::StackOrder);
}


std::string HelixPso::PrintStemlist(){
  string s=string();
  for(size_t i=0;i<HelixPso::stemlist.size();i++) s+=Str((int)i)+":"+RNAParameters::Print(HelixPso::stemlist[i])+"\n";
  return s;
}


void HelixPso::GetChromosomes(){
  chromosomes=std::vector<std::pair<double,std::vector<int> > >();
  //cout<<"Initial Chromosomes\n"<<std::endl;
  for(int i=0;i<population;i++) {
    if(i<start_idx || i>=stop_idx) {
      chromosomes.push_back(std::pair<double,std::vector<int> >());//NO_ENERGY_IMPROVEMENT,std::vector<int>()));
      //  cout<<" use dummy for chromosome "+Str(i)<<endl;
      continue;
    }
     vector<int> c=HelixPso::GetRandomPermutation(stemlist.size());
     std::pair<double,std::string> struc=Evaluate(c);
     //     double val=Evaluate(c).first;
     chromosomes.push_back(std::pair<double,std::vector<int> >(struc.first,c));
     nodes.push_back(HelixPso::ChromosomeToNode(c,stemlist));
     //    std::cout<<struc.second+" "+Str(struc.first)<<std::endl;
   }

}


 
void HelixPso::GetInitialPopulation(){
  changed=vector<vector<bool> >();
   for(int i=0;i<population;i++) {
     changed.push_back(vector<bool>(sequence.size(),false));
   }
   SelectEliteSolution();
   StoreBestSolution();  
 }


std::pair<double,std::string>  HelixPso::Evaluate(std::vector<int> chromosome){
  return Evaluate(HelixPso::ChromosomeToNode(chromosome,HelixPso::stemlist));
}


std::pair<double,std::string>  HelixPso::Evaluate(std::vector<std::pair<int,int> > node){
    std::pair<double,std::string> struc=RNAParameters::Evaluate(sequence,node);
    //incorporate a fraction of the centroid prediction quality into the score
    double mfe_component=(1-centroid_weight)*max(0,struc.first/p->mfe);
    double centroid_component=centroid_weight*(double)CentroidMatchBpCount(struc.second).first/(double)bp_centroid.size();
    double score=mfe_component+centroid_component;
    //  if(score>1) {
    //     cout<<"score "+Str(score)+struc.second<<std::endl;

    //   }
    struc.first=score;
    return struc;    //  return RNAParameters::Evaluate(sequence,node);
}




std::vector<std::pair<int,int> > HelixPso::ChromosomeToNode(int i){
  return HelixPso::ChromosomeToNode(chromosomes[i].second,HelixPso::stemlist);
}



int HelixPso::FindBestFitnessIdx(const std::vector<std::pair<double,std::vector<int> > >& chromosomes){
  double best=-NO_ENERGY_IMPROVEMENT;
  int best_idx=-1;
  for(int i=start_idx;i<stop_idx;i++){
    //  if(debug)cout<<Str(i)+" "+Str(chromosomes[i].first)<<endl;
    if(HasBetterFitness(chromosomes[i].first,best)) {
      best=chromosomes[i].first;
      best_idx=i;
    }
  }
  return best_idx;
}

/**
   Traverses the fitness vector, and returns index with
   worst fitness.Only to be called after EvaluateChromosomes()
*/ 

int HelixPso::FindWorstFitnessIdx(std::vector<std::pair<double,std::vector<int> > > chromosomes){
  //start at -infty so the first value is an improvement
  double worst=NO_ENERGY_IMPROVEMENT;
  int worst_idx=-1;
  for(int i=start_idx;i<stop_idx;i++){
    if(HasBetterFitness(worst,chromosomes[i].first)) {
      worst=chromosomes[i].first;
      worst_idx=i;
    }
  }
  return worst_idx;
}


bool HelixPso::HasBetterFitness(double f1,double f2){
 return (f1>f2);
 // return (f1<f2);
}


/**
   src and tgt are permutations. i is the first index where src[i]!=tgt[i].
   Then swap the indices l=src[i] and m=tgt[i] in src.
*/

void HelixPso::Mutate(std::vector<int>& off){
  int idx1=HelixPso::RandInt(off.size());
    int idx2=HelixPso::RandInt(off.size());
    //    cout<<" mutate index "+Str(idx1)+" and "+Str(idx2)+"\n";
    int val1=off[idx1];
    int val2=off[idx2];
    off[idx1]=val2;
    off[idx2]=val1;
     if(HelixPso::transposition_count_sections>0){
      HelixPso::transposition1st_counter[transposition_count_section][idx1]++;
      HelixPso::transposition2nd_counter[transposition_count_section][idx2]++;
    }
     //cout<<" mutate returns "<<endl;
}

  
bool HelixPso::MoveTowards(vector<int>& src,vector<int> tgt,vector<bool>& changed){

  int i=0;
  //skip further if on eof teh following is satisfied:
  //1.src and tgt have ame value
  //2.this value has already been changed in src
  //3.The value of has been exceeded, i.e. the current idx is not overwritten
  while( src[i]==tgt[i] || changed[i] ||  HelixPso::RandDouble()>HelixPso::swap_select) {
    i++;
    if(i==(int)src.size()) return false;//src;
  }
  //now:src[i]!=tgt[i]
  if(s->search_method=="cycle"){
    std::map<int,int> m=map<int,int>();
    //int  HelixPso::DoCycle(std::vector<int> p1,std::vector<int> p2,int idx,std::vector<int>& ret){
  while(true){
    m[i]=src[i];
    //cout<<"ret becomes:"+Print(ret)+"\n";
    //find index of tgt[i] in src;
    vector<int>::iterator it=find(src.begin(),src.end(),tgt[i]);
    i=it-src.begin();
    //cout<<"new i:"+Str(i)+"\n";
    //src[i] is now what tgt[i] was before the update which is the condition to continue the cycle
    bool i_covered=false;
    for(map<int,int>::iterator it=m.begin();it!=m.end();it++){
      if(it->second==src[i]) {
        i_covered=true;
        break;
      }
    }
    if(i_covered) break;
      //    if(find(m.begin(),ret.end(),src[i])!=ret.end()) break;
  } 
    for(map<int,int>::iterator it=m.begin();it!=m.end();it++){
      src[it->first]=it->second;  
    }
   changed[i]=true;
  }
  else{
      vector<int>::iterator it1=src.begin()+i;
      vector<int>::iterator it2=find(src.begin(),src.end(),tgt[i]);
      //      int j=it2-difference(src.begin(),it2);
      int x1=*it1;
      int x2=*it2;
      *it1=x2;
      *it2=x1; 
      if(HelixPso::transposition_count_sections>0){
	vector<int>::iterator distance_counter=src.begin();
        int j=0;
        while(distance_counter!=it2){
          j++;
          distance_counter++;
	}
       //std::cout<<"1st:"+Str(i)+"\n";
       //std::cout<<"2nd:"+Str(j)+"\n";
       //std::cout<<"transposition1st_counter size:"+Str(transposition1st_counter.size())+"\n";
       //std::cout<<"transposition1st_counter[0] size:"+Str(transposition1st_counter[0].size())+"\n";
       HelixPso::transposition1st_counter[transposition_count_section][i]++;
       HelixPso::transposition2nd_counter[transposition_count_section][j]++;
      }
      changed[i]=true;
  }
  return true;
}


void HelixPso::SetStartAndStopIdx(){
  start_idx=0;
  stop_idx=population;
}


//double alpha=0.95;
double subchain_length=1800;
double T_0=500;//5500;


bool HelixPso::AcceptTransition(std::string str_new,std::string str_old){

  bool debug=false;//true;//false;
double e_new=energy_of_struct(const_cast<char*>(p->RNA.c_str()),const_cast<char*>(str_new.c_str()));
double e_old=energy_of_struct(const_cast<char*>(p->RNA.c_str()),const_cast<char*>(str_old.c_str()));


  double delta_G=e_new-e_old;
  if(delta_G<0) return true;
  if(debug)  cout<<"delta_G "+Str(delta_G)<<std::endl;
  int T_steps=(int)((double)solution_count/(double)subchain_length);
  if(debug) cout<<"Tsteps "+Str(T_steps)<<std::endl;

 if(debug) cout<<"alpha "+Str(alpha)<<std::endl;
  double T=T_0*pow(alpha,T_steps);
  if(debug) cout<<"T at "+Str(solution_count)+" solutions: "+Str(T)<<std::endl;

  double p_accept=exp(-delta_G/T);
   if(debug)  cout<<"p_accept: "+Str(p_accept)<<std::endl;
  if(HelixPso::RandDouble()<=p_accept) return true;
  else return false;



}


/*
  Does in fact execute as many steps as prescribed by steps without 'looking'.
  Therefore, only the final solution counts as a solution. This makes
  DoNonGreedyIteration generate much less solution_count than DoGreedyIteration, due to
  the fact that DoGreedyIteration 'looks' after each step and decides to then keep
  or drop the result, hence producing one solution per step.
*/

void HelixPso::DoNonGreedyIteration(){
  bool debug=false;
  //  if(debug) cout<<"DoNonGreedyIteration "<<endl;
 

   for(int i=start_idx;i<stop_idx;i++){
        if(debug) cout<<"DoNonGreedyIteration "+Str(i)<<endl;
       //which indices in the permutation in a chromosome have changed
       changed[i]=vector<bool>(GetChromosomeSize(),false);
       int steps=s->steps_per_turn;//(int)min(GetTotalProbability(SV[i]),s->steps_per_turn);
       vector<int> new_val=chromosomes[i].second;
       for(int j=0;j<steps;j++){
          if(debug) cout<<" step "+Str(j)<<endl;
	 if(HelixPso::RandDouble()<=s->p_random_move){
            if(debug) cout<<" mutate"<<endl;
            Mutate(new_val);//Mutate2(new_val,idx1max);
	 }
	 else{
           // if(debug) cout<<" GetRandomReference"<<endl;
           std::vector<int> reference=GetRandomReference(i-start_idx);
	   //if(debug) cout<<" GOT RandomReference"<<endl;
           if(reference==new_val) {
             if(debug) cout<<" mutate as reference is the same as chrmosome"<<endl;
	     Mutate(new_val);
	   }
           else {
               if(debug)  cout<<" MoveTowards "<<endl;
	        if(!MoveTowards(new_val,GetRandomReference(i-start_idx),changed[i])) {
             if(debug)  cout<<" mutate"<<endl;
              Mutate(new_val);  
		}
	   }
           //new_val=HelixPso::CycleXover2(new_val,GetRandomReference(i));
	 }
       }
       //dont evaluate if nothing has changed
       if(new_val!=chromosomes[i].second ){
	 //        cout<<"there was a change in permutations of length "+Str((int)new_val.size())+" and "+Str((int)chromosomes[i].second.size())<<endl;
         std::vector<std::pair<int,int> > new_node=HelixPso::ChromosomeToNode(new_val,HelixPso::stemlist);
         if(new_node!=nodes[i]){

           //sol_generated_count++;
	   //cout<<sol_generated_count<<std::endl;
	   // cout<<"These result in different basepairs with count "+Str((int)new_node.size())+" and "+Str((int)nodes[i].size())<<endl;
	  //std::pair<double,std::string>  HelixPso::Evaluate(std::vector<int> chromosome){
	   std::pair<double,std::string> new_eval=Evaluate(new_node);
          double new_fitness=new_eval.first;
	  //double new_fitness=Evaluate(new_val);
          double old_fitness=chromosomes[i].first;
	  // std::cout<<"Chromosome "+Str(i)+" went from "+Str(old_fitness)+" to "+Str(new_fitness)<<std::endl;

	  //std::string RNAParameters::Structure(int length, std::vector<std::pair<int,int> > pairs){


	  std::string str_old=p->Structure(p->RNA.size(),HelixPso::ChromosomeToNode(chromosomes[i].second,HelixPso::stemlist));
          
	  if( (alpha!=0 && AcceptTransition(new_eval.second,str_old)) || (alpha==0 && HasBetterFitness(new_fitness,old_fitness))  ){

	       // if(HasBetterFitness(new_fitness,old_fitness) || HelixPso::RandDouble()<=0.2 ){
	    //std::cout<<"IMPROVED "<<std::endl;
       	    chromosomes[i]=std::pair<double,std::vector<int> >(new_fitness,new_val);
            nodes[i]=new_node;


             int correct=CorrectlyPredictedBpCount(new_eval.second).first;
             if(correct>record){
               cout<<" imprv "+new_eval.second+" "+Str(new_fitness)+" "+Str(correct)<<std::endl;
               record=correct;
 	    }

            

	    AddAcceptedSolution();
	  }
          //a produced result counts as a solution, even if it is not accepted as a chromosome
          else AddSolution();
       }
       else AddSolution();
       }
       if(HaveEnoughSolutions()) return;
   }
   if(debug)  cout<<"DoNonGreedyIteration returns"<<endl;
}

std::string HelixPso::PrintPermutation(std::vector<int> p){
  string s=string();
  for(size_t i=0;i<p.size();i++){
    s+=" "+Str(p[i]);
  }
  s+="\n";
  return s;
}


void HelixPso::AddChromosome(std::vector<int> permutation){
  chromosomes.push_back(std::pair<double,std::vector<int> >(Evaluate(permutation).first,permutation));
}



/**
   Memorize the best solution of the population just before another
   generation is done.
*/

void HelixPso::SelectEliteSolution(){
  int elite_idx=FindBestFitnessIdx(chromosomes);
  elite_solution=chromosomes[elite_idx];
}

/*
  Finds the best solution in new_chromomsomes and elite_solution and returns its fitness

*/

void  HelixPso::StoreBestSolution(){
  double elite_fitness=elite_solution.first;
  int current_best_fitness_idx=FindBestFitnessIdx(chromosomes);
  double current_best_fitness=chromosomes[current_best_fitness_idx].first;


  //   best_solution=make_pair(Evaluate(chromosomes[current_best_fitness_idx].second).first,chromosomes[current_best_fitness_idx].second);


  if(HasBetterFitness(elite_fitness,current_best_fitness)) {
    solutions_report.push_back(elite_fitness);
    //    best_solution=Evaluate(elite_solution.second).second;
    best_solution=make_pair(Evaluate(elite_solution.second).first,elite_solution.second);


    //  cout<<"Best solution :"+Str(elite_fitness)<<std::endl;
  }
  else {
    solutions_report.push_back(current_best_fitness);
    best_solution=make_pair(Evaluate(chromosomes[current_best_fitness_idx].second).first,chromosomes[current_best_fitness_idx].second);
    //    best_solution=Evaluate(chromosomes[current_best_fitness_idx].second).second;
    //   cout<<"Best solution :"+Str(current_best_fitness)<<std::endl;
  }

}


void HelixPso::DoIteration(){
  bool debug=false;//false;
  if(debug) cout<<" DoIteration"+Str(iterations_done)<<endl;
  //    cout<<"Global best "<<Str(gbest.back().first)<<std::endl;
   DoNonGreedyIteration();    
   //everyone looks whether one of their particles improved the gbest
   UpdatePbest();
  
   UpdateGbest();
   //  Elitism();
   //the clbest reference already needs to contain the chromosome exchanged by elitism.
   UpdateClbest();
    if(debug) cout<<"DoIteration calls  UpdateSV() "+Str(iterations_done)<<endl;
   UpdateSV();
   //DoSVStatistics();
   iterations_done++;
   if(debug) cout<<" Iteration done"<<endl;
}


void HelixPso::UpdateGbest(){
  LocalUpdateGbest();
}


void HelixPso::LocalUpdateGbest(){
  bool debug=false;
    if(debug) cout<<" LocalUpdateGbest calls FindBestFitnessIdx"<<endl;
   int idx=FindBestFitnessIdx(chromosomes);
   bool gbest_empty=true;
   if(gbest.size()>0) gbest_empty=false;
   if(!gbest_empty){
     double prev_gbest_energy=gbest[gbest.size()-1].first;
     if(HasBetterFitness(chromosomes[idx].first,prev_gbest_energy)) {
       gbest_changed=true;
       gbest.push_back(chromosomes[idx]); 
       std::pair<double,std::string>  gbest_ev=HelixPso::Evaluate(chromosomes[idx].second);
       matched=CorrectlyPredictedBpCount(gbest_ev.second).first;
       //cout<<"gbest "+gbest_ev.second+" "+Str(gbest_ev.first)+" "+Str(matched)<<std::endl;  
     }
     //ensures that gbest doesn't get worse if gbest is not written trough to chromosomes. To save communication between processors.
     else  gbest.push_back(gbest.back());
   }
   else gbest.push_back(chromosomes[idx]); 

   //HERE: Write through previous GBests unless outperformed by one of the current chromosomes.
   //saves having to send and resend the same Gbests that are not in the master's population.
     //   if(debug) 
     std::pair<double,std::string>  now_ev=Evaluate(chromosomes[idx].second);
     int correct=CorrectlyPredictedBpCount(now_ev.second).first;
      if(debug) cout<<" LocalUpdateGbest obtains "+now_ev.second+" "+Str(correct)+" and returns "<<endl;

}

std::pair<double,std::vector<int> >  HelixPso::FindBestChromosome( ){
    int idx=FindBestFitnessIdx(chromosomes);
    return chromosomes[idx];

}

void HelixPso::UpdateClbest(){
  for(int i=0;i<clusters;i++) LocalUpdateClbest(i);
}

bool HelixPso::LocalUpdateClbest(int idx){
   double prevBest=clbest[idx].back().first;

   bool debug=false;
    int start=idx*population/clusters;
    int end=(idx+1)*population/clusters;
     if(debug)cout<<" cluster "+Str(idx)+" start "+Str(start)+" and end "+Str(end)<<endl;
    //skip the clusters the processor is not involved in.
    //end is the starting index of teh next cluster end therefore moust be excepted for skipping
    if(end<=start_idx || start>=stop_idx) {
       if(debug)cout<<" skip as it is outside of range "+Str(start_idx)+" "+Str(stop_idx)<<endl;
       return false;
    }
    int bestIdx=start;
    for(int j=start+1;j<end;j++){
       if(debug) cout<<"index "+Str(j)+" "+Str(chromosomes[j].first)<<endl;
      if(HasBetterFitness(chromosomes[j].first,chromosomes[bestIdx].first)) {
	//         if(debug)cout<<"index "+Str(j)+" update from "+Str(chromosomes[bestIdx].first)+" to "+Str(chromosomes[j].first)<<endl;
         bestIdx=j;
      }
    }
    if(debug)cout<<"Best value for cluster "+Str(idx)+" "+Str(chromosomes[bestIdx].first)+" at idx "+Str(bestIdx)<<endl;
    clbest[idx].push_back(chromosomes[bestIdx]);
    double newBest=chromosomes[bestIdx].first;
    if(HasBetterFitness(newBest,prevBest)) return true;
    else return false;
}
int HelixPso::GetChromosomeSize(){
  return stemlist.size();
}

void HelixPso::Elitism(){ 
  int worstIdx=FindWorstFitnessIdx(chromosomes);
  chromosomes[worstIdx]=gbest[gbest.size()-1]; 
 
}

/**
   Increments the number of created solutions.Returns true, if 
   solution_count<total_solutions.That means that after the first call to AddSolution()
   that returned false, all further calls will returns false as well..
*/
bool HelixPso::HaveEnoughSolutions(){
   return (solution_count>=total_solutions);
}


void HelixPso::AddAcceptedSolution(){ 
  if(HelixPso::accepted_solution_count_sections>0){
    HelixPso::accepted_solution_counter[accepted_solution_count_section]++;

  }
  AddSolution();
}

void HelixPso::AddSolution(){
  solution_count++;
  if(solution_count%solutions_report_interval==0 || HaveEnoughSolutions()){
    
    // cout<<"solution_count is "+Str(solution_count)+", which is a multiple of "+Str(solutions_report_interval)+".\n";
    StoreBestSolution();   
   
    //cout<<"solutions_report has now size "+Str(solutions_report.size())+"\n";
  }
  if(HelixPso::transposition_count_sections>0){
    // bucket size for a section total_solutions/Node::transposition_count_sections;
    if(solution_count>=(transposition_count_section+1)*total_solutions/HelixPso::transposition_count_sections && transposition_count_section<HelixPso::transposition_count_sections) transposition_count_section++;   
  }
  if(HelixPso::accepted_solution_count_sections>0){
    // bucket size for a section total_solutions/Node::transposition_count_sections;
    if(solution_count>=(accepted_solution_count_section+1)*total_solutions/HelixPso::accepted_solution_count_sections  && accepted_solution_count_section<HelixPso::accepted_solution_count_sections) accepted_solution_count_section++;   
  }
 if(HelixPso::sv_stat_sections>0){
    // bucket size for a section total_solutions/Node::transposition_count_sections;
   if(solution_count>=(sv_stat_section+1)*total_solutions/HelixPso::sv_stat_sections && sv_stat_section<HelixPso::sv_stat_sections) {
     //     std::cout<<"solution_count :"+Str(solution_count)+"\n";
     DoSVStatistics();
     sv_stat_section++;   
  }
 }

}

std::vector<int> HelixPso::GetRandomReference(int i){
  bool debug=false;
   //sLog+="GetRandomReference:\n";
  if(debug) cout<<" GetRandomReference for particle "+Str(i)+" size of SV "+Str((int)SV.size())<<endl;

   std::map<std::pair<double,std::vector<int> >,double> m=SV[i];
   if(debug) cout<<" GetRandomReference gets total Probability"<<endl;
   double p=GetTotalProbability(m);
   double r=HelixPso::RandDouble()*p;
   //could do with some refactoring because very similar to std::vector<int> HelixPso::SelectChromosome(double total_fitness)
   int idx=0;
    if(debug)  cout<<" GetRandomReference traverese map for total p "+Str(r)<<endl;
   for(std::map<std::pair<double,std::vector<int> >,double>::iterator it=m.begin();it!=m.end();it++){
     ////sLog+=HelixPso::Print(it->first)+"\n";
      r-=it->second;
      if(r<=0) break;
      idx++; 
   }
   if(debug)  cout<<" GetRandomReference choose idx "+Str(idx)+" in reference vector "<<endl;
   std::map<std::pair<double,std::vector<int> >,double>::iterator it2=m.begin();
   advance(it2,idx);
   //sLog+="selected: "+HelixPso::Print(it2->first)+"\n";
   //the permuation of the reference chromosome
 if(debug)    cout<<" GetRandomReference returns "+PrintPermutation(it2->first.second)<<endl;
   return it2->first.second;
}


double HelixPso::GetTotalProbability( std::map<std::pair<double,std::vector<int> >,double> m){
  double p=0.0;
  for(std::map<std::pair<double,std::vector<int> >,double>::iterator it=m.begin();it!=m.end();it++) p+=it->second;
  return p;
}

void HelixPso::UpdatePbest(){
   for(int i=start_idx;i<stop_idx;i++){
     if(HasBetterFitness(chromosomes[i].first,pbest[i].first)) pbest[i]=chromosomes[i];
   }
  
}

void HelixPso::InitializeSV(){
  SV.clear();
  for(int i=start_idx;i<stop_idx;i++){
    vector<int> permutation=HelixPso::GetRandomPermutation(HelixPso::stemlist.size());
    std::pair<double,std::vector<int> > chromosome=std::pair<double,std::vector<int> >(Evaluate(permutation).first,permutation);
    std::map< std::pair<double,std::vector<int> >,double> m=std::map< std::pair<double,std::vector<int> >,double>();
    m[chromosome]=1.0;
    SV.push_back(m);
  }
}

// std::string HelixPso::PrintReferencePoints(std::map< std::pair<double,std::vector<int> >,double> m){
//    string s=string();
//    for(std::map< std::pair<double,std::vector<int> >,double>::iterator it=m.begin();it!=m.end();it++) {
//      s+=HelixPso::Print(it->first.second)+":"+Str(it->second)+"\n";
//    }
//    return s;
// }

void HelixPso::UpdateSV(){
  bool debug=false;
  if(debug)   cout<<" UpdateSV "<<endl;
  //   cout<<"evaporation_threshold:"+Str(HelixPso::evaporation_threshold)+"\n";    
  int particles_per_cluster=(int)(population/clusters);
  for(int i=start_idx;i<stop_idx;i++){
     std::map< std::pair<double,std::vector<int> >,double> m=SV[i-start_idx];
     //cout<<"Update SV["+Str(i)+"] from:\n";   
     ////sLog+=PrintReferencePoints(m);
    
     //evaporate
     // double evaporation_threshold=0.01;
     if(debug) cout<<" UpdateSV particle "+Str(i)+" with energy "+Str(chromosomes[i].first)<<endl;
     //     cout<<"map size "+m.size()<<std::endl;
     for(std::map< std::pair<double,std::vector<int> >,double>::iterator it=m.begin();it!=m.end();){//;it++){
       //for(std::map< std::pair<double,std::vector<int> >,double>::bidirectional_iterator it=m.rbegin();it!=m.rend();it++){
       m[it->first]*=s->evaporate;
       if(m.size()>1 && m[it->first]<HelixPso::evaporation_threshold) {
         m.erase(it++);
   //       m.erase(it);
//          cout<<"erase"<<std::endl;
//          cout<<"map size "+m.size()<<std::endl;
       }
       else ++it;//cout<<"Dont erase"<<std::endl;
     }
     //   std::map< std::pair<double,std::vector<int> >,double>::iterator it=remove_if(m.begin(),m.end(),std::bind2nd(BasePairConflict(),HelixPso::evaporation_threshold));


     if(debug)  cout<<" UpdateSV particle "+Str(i)+"half evaporated"<<endl;
      std::pair<double,std::vector<int> > pb=pbest[i];     
     double r1=HelixPso::RandDouble();
     if(m.find(pb)!=m.end()) m[pb]+= r1*s->p_bonus_pbest;
     else                    m[pb] = r1*s->p_bonus_pbest;

     if(debug)  cout<<" UpdateSV particle "+Str(i)+"evaporated"<<endl;
     
     ////sLog+="to:\n";
     ////sLog+=PrintReferencePoints(m);
     

     int cluster=(int)(i/particles_per_cluster);
     //rounding may lead to cluster=clusters, but the last cluster contains 'leftover' particles, so index becomes clusters-1
     if(cluster>=clusters) cluster=clusters-1;

     if(debug) {
        cout<<" cluster "+Str(cluster)+" has best energy "+Str(clbest[cluster].back().first)<<endl;
     }
     bool best_of_cluster=false;
     if(clbest[cluster].back()==chromosomes[i]) best_of_cluster=true;
      if(debug) cout<<" best_of_cluster "<<best_of_cluster<<endl;

      std::pair<double,std::vector<int> > gb= std::pair<double,std::vector<int> > ();
     if(best_of_cluster) gb=gbest.back();     
     else                gb=clbest[cluster].back();  
     double r2=HelixPso::RandDouble();   
     if(m.find(gb)!=m.end()) m[gb]+=r2*s->p_bonus_gbest;
     else                    m[gb] =r2*s->p_bonus_gbest;
    SV[i-start_idx]=m;
  }
   if(debug) cout<<" UpdateSV returns"<<endl;
}

void HelixPso::DoSVStatistics(){
  return;
   //do statistics on SV[i]
    if(HelixPso::sv_stat_sections>0){
      int particles_moved=solution_count-last_stat_update;//%(total_solutions/sv_stat_sections);
      //      std::cout<<"particles_moved:"+Str(particles_moved)+"\n";
      for(int i=0;i<particles_moved;i++){
      sv_length=0;   
      sv_mass=0.0;
      for(std::map< std::pair<double,std::vector<int> >,double>::iterator it=SV[i].begin();it!=SV[i].end();it++) {
	sv_mass+=SV[i][it->first];
	sv_length++;
      }
      sv_length_counter[sv_stat_section]+=sv_length;
      sv_mass_counter[sv_stat_section]+=sv_mass;
      //      HelixPso::sv_length_counter+=sv_length;
      //HelixPso::sv_mass_counter+=sv_mass;
      for(std::map< std::pair<double,std::vector<int> >,double>::iterator it=SV[i].begin();it!=SV[i].end();it++){
	double reference_mass=SV[i][it->first];
        for(size_t j=0;j<sv_mass_fractions.size();j++){
	  if(reference_mass>=sv_mass_fractions[j]*sv_mass) {
	    sv_distribution_counter[sv_stat_section][j]+=1.0;///(double)sv_length;
	    break;
	    }
	}
      }
    }
      last_stat_update=solution_count;
      //std::cout<<"last_stat_update:"+Str(last_stat_update)+"\n";
    }
}


    std::pair<int,int> HelixPso::CorrectlyPredictedBpCount(std::string structure){
      bool debug=false;
      if(debug) cout<<"CorrectlyPredictedBpCount native has "+Str((int)bp_native.size())+" bp"<<std::endl;

      int bpcorrect=0;
      std::vector<std::pair<int,int> > bpfound=RNAParameters::MakeBasePairList (structure);
      if(debug) cout<<"Argument structure has "+Str((int)bpfound.size())+" bp"<<std::endl;
      for(std::vector<std::pair<int,int> >::iterator it=bpfound.begin();it!=bpfound.end();it++){
        if(find(bp_native.begin(),bp_native.end(),*it)!= bp_native.end()) bpcorrect++;
      }
      return std::make_pair<int,int>(bpcorrect,bpfound.size());
    }




    std::pair<int,int> HelixPso::CentroidMatchBpCount(std::string structure){
      bool debug=false;


      int bpcorrect=0;
      std::vector<std::pair<int,int> > bpfound=RNAParameters::MakeBasePairList (structure);
      if(debug) cout<<"Argument structure has "+Str((int)bpfound.size())+" bp"<<std::endl;
      for(std::vector<std::pair<int,int> >::iterator it=bpfound.begin();it!=bpfound.end();it++){
        if(find(bp_centroid.begin(),bp_centroid.end(),*it)!= bp_centroid.end()) bpcorrect++;
      }
      return std::make_pair<int,int>(bpcorrect,bpfound.size());
    }

